#!/bin/bash

# set sane defaults
IFS=$' \t\n'
export LANG=C
export LC_NUMERIC=C
export LC_TIME=C

die()
{
  [ "$1" == '0' ] && printf '%s\n' "$2" || printf '%s\n' "$2" 1>&2
  exit "$1"
}

syntax()
{
  die "$1" \
"Usage: op5 mt report-data-reinsert -s <ts> [OPTION].. [--] [NODE]..

 -h, --help         This syntax help text.
 -i, --input-file   Disables collection mode. The file at the given path con-
                    tains the alert log entries to insert into the database.
 -o, --output-file  Enables collection mode, where alert logs are fetched from
                    all nodes, or the ones given on command line (+localhost).
                    The collected alert log entries will be stored in the file
                    at the given path.
 -s, --since        The database table (report_data) will be rewritten starting
                    at this time. This also affects which entries to collect.
                    A unix timestamp is expected initially, but any string
                    compatible with date(1)'s --date option is accepted.

If neither -i or -o is given, the tool defaults to setting -o to a temp file.

Before running this tool, it is recommended to:
* Cancel all current scheduled downtimes (or run the tool when there are none).
* Shut down Monitor on all nodes, and don't start it until finished processing.
"
}

confirm()
{
  read -p 'Continue (y/N)? ' -r -N1 q
  printf '\n'
  [ "$q" == 'y' -o "$q" == 'Y' ] || die '0' 'Aborting.'
}

nodetest()
{
  printf '\nVerifying node command execution capabilities...\n'
  for name in $names; do
    mon node rawctrl "$name" --test || exit $?
  done
}

nodeget_confirm()
{
  local ts="$(date -d"@$since" '+%Y-%m-%d %H:%M:%S')"
  echo -n "
Next up is:
1) Collect alert log data (since $ts) from the listed nodes.
2) Sort and deduplicate the data.
3) Write the results to file: $file

Please be advised that the sorting will create additional temporary files in
/tmp (or \$TMPDIR if set via environ), which might require large amounts of
disk space. This amount cannot be pre-determined.

The process might also be very time consuming, depending on the amount of data
to read at each node. Only the alert events found in the log data at each node
is downloaded (compressed) via the network (in case of remote nodes).

Shutting down Monitor at all listed nodes before continuing is recommended!

"
  confirm
}

nodeget() {
  local cmd lines name

  (
  # Enable pipefail, hence the subshell.
  set -o pipefail

  cmd="mon log merge -1 '$since' nagios | "
  cmd+="mon log range '$since' | "
  cmd+="mon log filter nagios/{{host,service}-{alert,downtime},startstop}"

  printf '\n'
  for name in $names; do
    printf '%s (%s) starting...\n' "$(date)" "$name" 1>&2
    SSH_OPTS_EXTRA='-C' mon node rawctrl "$name" "$cmd" | tee "$file.$name" || \
      die '1' "$(date) ($name) command failed: $cmd"
    printf '%s (%s) done (entries: %s) (errors: %s)\n' \
      "$(date)" "$name" \
      "$(grep '^\[' "$file.$name" | wc -l)" \
      "$(grep -v '^\[' "$file.$name" | wc -l)" 1>&2
  done | sort -u | grep '^\[' > "$file" || \
    die '1' "Error detected. Exiting."
  printf '\n'

  lines="$(cat "$file" | wc -l)" || exit $?
  printf 'Final number of collected log entries: %s\n' "$lines"
  ) || exit $?
}

mysqldo()
{
  out="$(mysql -BNr merlin -e "$1" 2>&1)" || \
    die '1' $'failed\n'"$out"

  [ -n "$out" ] && \
    die '1' $'unexpected output:\n'"$out"

  printf 'ok\n'
}

import_confirm()
{
  local ts="$(date -d@"$since" '+%Y-%m-%d %H:%M:%S')"
  echo -n "
Next up is:
1) Delete all current report data entries that are timestamped $ts or more recently.
2) Insert the report data entries found in '$file'

"
  confirm
}

import()
{
  printf '\nVerifying MySQL connectivity... '
  mysqldo 'select id from report_data limit 0'

  printf 'Deleting old report data entries... '
  mysqldo "delete from report_data where timestamp >= $since"

  printf '\n'
  mon log import --incremental="$since" --nagios-cfg=/dev/null "$file" || \
    exit $?
}


unset file infile outfile since

while [ "$#" -gt '0' ]; do
  case "$1" in
    -h|--help)
      syntax 0
      ;;
    -i|--input-file)
      [ -n "$2" ] || syntax 1
      infile="$(readlink -m -- "$2")"
      shift 2
      ;;
    -o|--output-file)
      [ -n "$2" ] || syntax 1
      outfile="$(readlink -m -- "$2")"
      shift 2
      ;;
    -s|--since)
      [ -n "$2" ] || syntax 1
      since="$2"
      shift 2
      ;;
    --)
      shift
      break
      ;;
    *)
      syntax 1
      ;;
  esac
done

if [ -z "$since" ]; then
  syntax 1
elif ! [ "$since" -ge '0' ] &> /dev/null; then
  since="$(date -d "$since" +%s)" || exit $?
fi

[ -n "$infile" -a -n "$outfile" ] && \
  die '1' 'ERROR: Arguments -i and -o are mutually exclusive.'

ssh='yes'
if [ -n "$outfile" ]; then
  if [ -f "$outfile" -a -s "$outfile" ]; then
    printf 'File (%s) already exists and will be overwritten.' "$outfile"
    confirm
  fi
  : > "$outfile" || exit $?
  file="$outfile"
elif [ -n "$infile" ]; then
  file="$infile"
  ssh=''
else
  file="$(mktemp -t "alerts.$since.$(date +%s).XXXXXX.log")" || \
    die '1' "ERROR: Failed creating temporary file ($file)."$'\n'"$file"
fi


if [ -n "$ssh" ]; then
  if [ -n "$1" ]; then
    names="$HOSTNAME $*"
  else
    names="$HOSTNAME $(mon node list)"
  fi

  for name in $names; do
    : > "$file.$name" || \
      die '1' "ERROR: Failed creating temporary file ($file.$name)."
    chmod 600 "$file.$name" || \
      die '1' "ERROR: Failed setting permissions of temporary file ($file.$name)."
  done

  nodetest
  nodeget_confirm
  nodeget
else
  [ -n "$1" ] && \
    die '1' 'ERROR: Not expecting node names in combination with -i argument.'
fi

import_confirm
import

printf '\nAll done!\n'
exit 0
